"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const performance_now_1 = __importDefault(require("performance-now"));
const common_1 = require("./common");
function createAutometricMiddleware(name, baseOptions = {}) {
    const register = common_1.createRegistry();
    const { labels = {}, registers = [] } = baseOptions;
    registers.push(register);
    baseOptions.rewriteLabels = baseOptions.rewriteLabels || ((passLabels) => passLabels);
    const metrics = {
        calls: common_1.createCounter({
            help: `Autometric Counter for requests on the "${name}" Route`,
            name: name + "_calls_total",
            registers,
        }),
        durations: common_1.createSummary({
            help: `Autometric Summary for the durations on the "${name}" Route`,
            name: name + "_durations_ms",
            registers,
        }),
        fails: common_1.createCounter({
            help: `Autometric Counter for unsuccessfully respond requests of the "${name}" Route`,
            name: name + "_fails_total",
            registers,
        }),
        incomingThroughput: common_1.createSummary({
            help: `Autometric Summary for the incoming throughput in bytes of the "${name}" Route`,
            name: name + "_incoming_bytes",
            registers,
        }),
        outgoingThroughput: common_1.createSummary({
            help: `Autometric Summary for the outgoing throughput in bytes of the "${name}" Route`,
            name: name + "_outgoing_bytes",
            registers,
        }),
        successes: common_1.createCounter({
            help: `Autometric Counter for successfully respond requests of the "${name}" Route`,
            name: name + "_successes_total",
            registers,
        }),
    };
    return {
        createMiddleware(options = {}) {
            return (request, response, next) => {
                const startTimestamp = performance_now_1.default();
                const startDate = new Date();
                let closed = false;
                let currentLabels = Object.assign({}, labels, options.labels);
                if (request /* Express response */.method && baseOptions.addMethodLabel) {
                    currentLabels.method = request.method;
                }
                metrics.calls.inc(currentLabels, 1, startDate);
                function closeConnection() {
                    const stopTimestamp = performance_now_1.default();
                    const endDate = new Date();
                    if (closed) {
                        return;
                    }
                    // Safty goes first and first come - first service. Sometimes close get called multiple times,
                    // or finish and time.
                    closed = true;
                    if (baseOptions.addStatusCodeLabel) {
                        currentLabels.statusCode = response.statusCode.toString();
                    }
                    currentLabels = baseOptions.rewriteLabels(currentLabels, baseOptions, { request, response });
                    metrics.durations.observe(currentLabels, stopTimestamp - startTimestamp);
                    if (response.statusCode >= 400) {
                        metrics.fails.inc(currentLabels, 1, endDate);
                    }
                    else {
                        metrics.successes.inc(currentLabels, 1, endDate);
                    }
                    metrics.incomingThroughput.observe(currentLabels, request.socket.bytesRead);
                    metrics.outgoingThroughput.observe(currentLabels, request.socket.bytesWritten);
                }
                response.on("finish", () => {
                    if (baseOptions.addClosedByLabel) {
                        currentLabels.closedBy = "server";
                    }
                    closeConnection();
                });
                response.on("close", () => {
                    if (baseOptions.addClosedByLabel) {
                        currentLabels.closedBy = "client";
                    }
                    closeConnection();
                });
                next();
            };
        },
        labels,
        prefix: name,
        register,
    };
}
exports.createAutometricMiddleware = createAutometricMiddleware;
