"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chai_1 = require("chai");
const freeport_1 = __importDefault(require("freeport"));
const http_1 = require("http");
const prom_client_1 = require("prom-client");
const request_promise_1 = __importDefault(require("request-promise"));
const _1 = require("./");
describe("AutometricMiddleware", () => {
    const requestResponseMock = {
        on(name, callback) { callback(); },
        method: "GET",
        socket: {
            bytesRead: 111,
            bytesWritten: 333,
        },
        statusCode: 200,
    };
    describe("creation of the middleware object", () => {
        let middlewareObject;
        const prefix = "instance_of_middleware";
        const labels = { test: "OK" };
        before(() => {
            _1.autometricRegister.getSingleMetric("autometric_counters_total").reset();
            _1.autometricRegister.getSingleMetric("autometric_registers_total").reset();
            _1.autometricRegister.getSingleMetric("autometric_summaries_total").reset();
        });
        before(() => {
            middlewareObject = _1.createAutometricMiddleware(prefix, { labels });
        });
        it("should have a metrics register", () => void chai_1.expect(middlewareObject.register).instanceOf(Object));
        it("should have the prefix", () => void chai_1.expect(middlewareObject.prefix).equal(prefix));
        it("should have the labels object", () => void chai_1.expect(middlewareObject.labels).deep.equal(labels));
        it("should return a middleware function", () => void chai_1.expect(middlewareObject.createMiddleware()).instanceOf(Function));
        it("should count the creation of a counters", () => void chai_1.expect(_1.autometricRegister.getSingleMetric("autometric_counters_total").get().values[0].value).equal(3));
        it("should count the creation of a register", () => void chai_1.expect(_1.autometricRegister.getSingleMetric("autometric_registers_total").get().values[0].value).equal(1));
        it("should count the creation of a summaries", () => void chai_1.expect(_1.autometricRegister.getSingleMetric("autometric_summaries_total").get().values[0].value).equal(3));
        it("should have the calls_total Counter in the register", () => {
            chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_calls_total")).instanceOf(prom_client_1.Counter);
        });
        it("should have the durations_ms Summary in the register", () => {
            chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_durations_ms")).instanceOf(prom_client_1.Summary);
        });
        it("should have the fails_total Counter in the register", () => {
            chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_fails_total")).instanceOf(prom_client_1.Counter);
        });
        it("should have the incoming_bytes Summary in the register", () => {
            chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_incoming_bytes")).instanceOf(prom_client_1.Summary);
        });
        it("should have the outgoing_bytes Summary in the register", () => {
            chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_outgoing_bytes")).instanceOf(prom_client_1.Summary);
        });
        it("should have the successes_total Counter in the register", () => {
            chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_successes_total")).instanceOf(prom_client_1.Counter);
        });
    });
    describe("work as a middleware", () => {
        let middlewareObject;
        const prefix = "instance_of_middleware";
        const labels = { test: "OK" };
        before(() => {
            middlewareObject = _1.createAutometricMiddleware(prefix, { labels });
        });
        it("should pass to next", (done) => {
            middlewareObject.createMiddleware()(requestResponseMock, requestResponseMock, () => done());
        });
    });
    describe("counters", () => {
        describe("calls", () => {
            let middlewareObject;
            const prefix = "calls_total_counters";
            before(() => {
                middlewareObject = _1.createAutometricMiddleware(prefix, {
                    addMethodLabel: true,
                    addStatusCodeLabel: true,
                });
            });
            before((done) => {
                middlewareObject.createMiddleware()(requestResponseMock, requestResponseMock, done);
            });
            it("should count a number of calls of the middleware", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_calls_total")
                    .get().values[1].value).equal(1);
            });
            it("should have the given method", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_calls_total")
                    .get().values[1].labels.method).equal("GET");
            });
            it("should have the given status code", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_calls_total")
                    .get().values[1].labels.statusCode).equal("200");
            });
        });
        describe("fails", () => {
            let middlewareObject;
            const prefix = "fails_total_counters";
            before(() => {
                middlewareObject = _1.createAutometricMiddleware(prefix, {
                    addMethodLabel: true,
                    addStatusCodeLabel: true,
                });
            });
            before((done) => {
                middlewareObject.createMiddleware()(requestResponseMock, Object.assign({}, requestResponseMock, { statusCode: 500 }), done);
            });
            it("should count a number of calls of the middleware", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_fails_total")
                    .get().values[1].value).equal(1);
            });
            it("should have the given method", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_fails_total")
                    .get().values[1].labels.method).equal("GET");
            });
            it("should have the given status code", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_fails_total")
                    .get().values[1].labels.statusCode).equal("500");
            });
        });
        // describe("bytes incoming", () => {});
        // describe("bytes outgoing", () => {});
        describe("successes", () => {
            let middlewareObject;
            const prefix = "success_total_counters";
            before(() => {
                middlewareObject = _1.createAutometricMiddleware(prefix, {
                    addMethodLabel: true,
                    addStatusCodeLabel: true,
                });
            });
            before((done) => {
                middlewareObject.createMiddleware()(requestResponseMock, requestResponseMock, done);
            });
            it("should count a number of calls of the middleware", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_successes_total")
                    .get().values[1].value).equal(1);
            });
            it("should have the given method", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_successes_total")
                    .get().values[1].labels.method).equal("GET");
            });
            it("should have the given status code", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_successes_total")
                    .get().values[1].labels.statusCode).equal("200");
            });
        });
    });
    describe("integration test", () => {
        context("success", () => {
            let port;
            let server;
            let middlewareObject;
            const prefix = "int_success";
            before(() => {
                middlewareObject = _1.createAutometricMiddleware(prefix, {
                    addClosedByLabel: true,
                    addMethodLabel: true,
                    addStatusCodeLabel: true,
                });
            });
            after((done) => {
                server.close(done);
            });
            before((done) => {
                freeport_1.default((err, result) => {
                    port = result;
                    server = http_1.createServer((req, res) => {
                        middlewareObject.createMiddleware()(req, res, () => {
                            res.writeHead(200);
                            res.end("Test-Response");
                        });
                    });
                    server.listen(port, done);
                });
            });
            before(() => Promise.all([
                request_promise_1.default("http://localhost:" + port + "/andsix"),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
            ]));
            it("it should count a call", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_calls_total")
                    .get().values.map((value) => value.value).reduce((p, c) => p + c, 0))
                    .equal(10);
            });
            it("it should count the success", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_successes_total")
                    .get().values.map((value) => value.value).reduce((p, c) => p + c, 0))
                    .equal(10);
            });
            it("it should count the fails", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_fails_total")
                    .get().values.map((value) => value.value).reduce((p, c) => p + c, 0))
                    .equal(0);
            });
            it("it should summarize the incoming bytes", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_incoming_bytes")
                    .hashMap["closedBy:server,method:GET,statusCode:200"].sum)
                    .equal(606);
            });
            it("it should summarize the outgoing bytes", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_outgoing_bytes")
                    .hashMap["closedBy:server,method:GET,statusCode:200"].sum)
                    .equal(1260);
            });
        });
        context("fail", () => {
            let port;
            let server;
            let middlewareObject;
            const prefix = "int_fails";
            before(() => {
                middlewareObject = _1.createAutometricMiddleware(prefix, {
                    addClosedByLabel: true,
                    addMethodLabel: true,
                    addStatusCodeLabel: true,
                });
            });
            after((done) => {
                server.close(done);
            });
            before((done) => {
                freeport_1.default((err, result) => {
                    port = result;
                    server = http_1.createServer((req, res) => {
                        middlewareObject.createMiddleware()(req, res, () => {
                            res.writeHead(500);
                            res.end("Test-Response");
                        });
                    });
                    server.listen(port, done);
                });
            });
            before(() => Promise.all([
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
                request_promise_1.default("http://localhost:" + port),
            ]).catch(() => { }));
            it("it should count a call", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_calls_total")
                    .get().values.map((value) => value.value).reduce((p, c) => p + c, 0))
                    .equal(3);
            });
            it("it should count the success", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_successes_total")
                    .get().values.map((value) => value.value).reduce((p, c) => p + c, 0))
                    .equal(0);
            });
            it("it should count the fails", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_fails_total")
                    .get().values.map((value) => value.value).reduce((p, c) => p + c, 0))
                    .equal(3);
            });
            it("it should count the incoming bytes", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_incoming_bytes")
                    .hashMap["closedBy:server,method:GET,statusCode:500"].sum)
                    .equal(180);
            });
            it("it should count the outgoing bytes", () => {
                chai_1.expect(middlewareObject.register.getSingleMetric(prefix + "_outgoing_bytes")
                    .hashMap["closedBy:server,method:GET,statusCode:500"].sum)
                    .equal(435);
            });
        });
    });
});
