"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const performance_now_1 = __importDefault(require("performance-now"));
const stream_1 = require("stream");
const common_1 = require("./common");
function createAutometricStreamPipe(name, baseOptions = {}) {
    const register = common_1.createRegistry();
    const { registers = [] } = baseOptions;
    registers.push(register);
    const metrics = {
        chunkSizes: common_1.createSummary({
            help: `Autometric Summary for the chunk-size of the "${name}" Stream`,
            name: name + "_chunk_sizes_bytes",
            registers,
        }),
        durations: common_1.createSummary({
            help: `Autometric Summary for the durations of the "${name}" Stream`,
            name: name + "_durations_ms",
            registers,
        }),
        elapsedTime: common_1.createSummary({
            help: `Autometric Summary for the elapsed time between emitted chunks of the "${name}" Stream`,
            name: name + "_elapsed_time_ms",
            registers,
        }),
        ends: common_1.createCounter({
            help: `Autometric Counter for ended Streams of "${name}"`,
            name: name + "_ends_total",
            registers,
        }),
        nonEmits: common_1.createCounter({
            help: `Autometric Counter for Streams of "${name}" without emitting data`,
            name: name + "_non_emits_total",
            registers,
        }),
        throughput: common_1.createSummary({
            help: `Autometric Summary for the throughput in bytes of the "${name}" Stream`,
            name: name + "_throughput_bytes",
            registers,
        }),
    };
    return _a = class AutometricStreamPassThrough extends stream_1.PassThrough {
            constructor(options = {}) {
                let firstTimestamp = performance_now_1.default(); // set already now a value to prevent bugs with non-emitting streams
                options.labels = options.labels || {};
                options.rewriteLabels = options.rewriteLabels || ((passLabels) => passLabels);
                let currentLabels = Object.assign({}, AutometricStreamPassThrough.labels, options.labels);
                let firstChunk = false;
                let throughput = 0;
                let previousTimestamp;
                const transform = (chunk, encoding, callback) => {
                    const startTimestamp = performance_now_1.default();
                    currentLabels = options.rewriteLabels(currentLabels, chunk, encoding, baseOptions, options);
                    if (firstChunk) {
                        metrics.elapsedTime.observe(currentLabels, startTimestamp - previousTimestamp);
                    }
                    else {
                        firstChunk = true;
                        firstTimestamp = startTimestamp;
                    }
                    if (chunk instanceof Buffer || typeof chunk === "string") {
                        metrics.chunkSizes.observe(currentLabels, chunk.length);
                        throughput += chunk.length;
                    }
                    else {
                        metrics.chunkSizes.observe(currentLabels, 1);
                        throughput += 1;
                    }
                    previousTimestamp = performance_now_1.default();
                    callback(undefined, chunk);
                };
                super({
                    objectMode: true,
                    transform,
                });
                this.on("end", () => {
                    const endTimestamp = performance_now_1.default();
                    const endDate = new Date();
                    if (!firstChunk) {
                        metrics.nonEmits.inc(currentLabels, 1, endDate);
                    }
                    metrics.ends.inc(currentLabels, 1, endDate);
                    metrics.durations.observe(currentLabels, endTimestamp - firstTimestamp);
                    metrics.throughput.observe(currentLabels, throughput);
                });
            }
        },
        _a.register = register,
        _a.prefix = name,
        _a.labels = baseOptions.labels || {},
        _a;
    var _a;
}
exports.createAutometricStreamPipe = createAutometricStreamPipe;
