"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const chai_1 = require("chai");
const prom_client_1 = require("prom-client");
const stream_1 = require("stream");
const _1 = require("./");
const index_1 = require("./index");
class Pipe extends stream_1.PassThrough {
    constructor({ objectMode = false } = {}) {
        super({ objectMode, transform: (chunk, encoding, next) => { next(undefined, chunk); } });
    }
}
describe("AutometricStream", () => {
    describe("creation of the Pass-Through class", () => {
        it("should return an instance of a Promise", () => {
            chai_1.expect(new (_1.createAutometricStreamPipe("instance_of_pass_through"))()).instanceOf(stream_1.PassThrough);
        });
        describe("static properties", () => {
            let AutometricStreamPipe;
            const name = "static_properties";
            const labels = { label: "just-a-test" };
            before(() => {
                index_1.autometricRegister.getSingleMetric("autometric_counters_total").reset();
                index_1.autometricRegister.getSingleMetric("autometric_registers_total").reset();
                index_1.autometricRegister.getSingleMetric("autometric_summaries_total").reset();
            });
            before(() => {
                AutometricStreamPipe = _1.createAutometricStreamPipe(name, {
                    labels,
                });
            });
            it("should have the prefix as prefix attribute", () => {
                chai_1.expect(AutometricStreamPipe.prefix).to.equal(name);
            });
            it("should have the labels as labels attribute", () => {
                chai_1.expect(AutometricStreamPipe.labels).to.deep.equal(labels);
            });
            it("should have the chunk-sizes summary in the register", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")).instanceOf(prom_client_1.Summary);
            });
            it("should have the durations summary in the register", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_durations_ms")).instanceOf(prom_client_1.Summary);
            });
            it("should have the throughput summary in the register", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_elapsed_time_ms")).instanceOf(prom_client_1.Summary);
            });
            it("should have the ends counter in the register", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_ends_total")).instanceOf(prom_client_1.Counter);
            });
            it("should have the non-emits counter in the register", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_non_emits_total")).instanceOf(prom_client_1.Counter);
            });
            it("should have the throughput summary in the register", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_throughput_bytes")).instanceOf(prom_client_1.Summary);
            });
            it("should count the creation of a counters", () => void chai_1.expect(index_1.autometricRegister.getSingleMetric("autometric_counters_total").get().values[0].value).equal(2));
            it("should count the creation of a register", () => void chai_1.expect(index_1.autometricRegister.getSingleMetric("autometric_registers_total").get().values[0].value).equal(1));
            it("should count the creation of a summaries", () => void chai_1.expect(index_1.autometricRegister.getSingleMetric("autometric_summaries_total").get().values[0].value).equal(4));
        });
    });
    describe("work as a PassThrough stream", () => {
        let AutometricStreamPipe;
        const name = "as_stream";
        before(() => {
            AutometricStreamPipe = _1.createAutometricStreamPipe(name);
        });
        it("should pass data", (done) => {
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.once("data", (data) => {
                chai_1.expect(data.toString()).equal("Just a test");
                done();
            });
            inPipe.write("Just a test");
        });
        it("should pass data in object mode", (done) => {
            const inPipe = new Pipe({ objectMode: true });
            const outPipe = new Pipe({ objectMode: true });
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.once("data", (data) => {
                chai_1.expect(data).deep.equal({ message: "Just a test" });
                done();
            });
            inPipe.write({ message: "Just a test" });
        });
    });
    describe("counters", () => {
        let AutometricStreamPipe;
        const name = "counters";
        before(() => {
            AutometricStreamPipe = _1.createAutometricStreamPipe(name);
        });
        it("should count an emitted chunk", () => {
            AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            inPipe.write("Just a test");
            chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")
                .get().values.reduce((prev, elem) => {
                /* istanbul ignore if */
                if (prev) {
                    return prev;
                }
                if (elem.metricName === name + "_chunk_sizes_bytes_count") {
                    return elem.value;
                }
            }, null)).equal(1);
        });
        it("should summarize the chunk sizes", () => {
            AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            const testStr = "Just a test";
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            inPipe.write(testStr);
            chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")
                .get().values.reduce((prev, elem) => {
                if (prev) {
                    return prev;
                }
                if (elem.metricName === name + "_chunk_sizes_bytes_sum") {
                    return elem.value;
                }
            }, null)).equal(testStr.length);
        });
        it("should summarize the chunk sizes of strings in object mode", () => {
            AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes").reset();
            const inPipe = new Pipe({ objectMode: true });
            const outPipe = new Pipe({ objectMode: true });
            const autometricStream = new AutometricStreamPipe();
            const testStr = "Just a test";
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            inPipe.write(testStr);
            chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")
                .get().values.reduce((prev, elem) => {
                if (prev) {
                    return prev;
                }
                if (elem.metricName === name + "_chunk_sizes_bytes_sum") {
                    return elem.value;
                }
            }, null)).equal(testStr.length);
        });
        it("should summarize the chunk sizes of buffers in object mode", () => {
            AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes").reset();
            const inPipe = new Pipe({ objectMode: true });
            const outPipe = new Pipe({ objectMode: true });
            const autometricStream = new AutometricStreamPipe();
            const testBuffer = Buffer.from("Just a test");
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            inPipe.write(testBuffer);
            chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")
                .get().values.reduce((prev, elem) => {
                if (prev) {
                    return prev;
                }
                if (elem.metricName === name + "_chunk_sizes_bytes_sum") {
                    return elem.value;
                }
            }, null)).equal(testBuffer.length);
        });
        it("should summarize the number of chunks in pure object mode", () => {
            AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes").reset();
            const inPipe = new Pipe({ objectMode: true });
            const outPipe = new Pipe({ objectMode: true });
            const autometricStream = new AutometricStreamPipe();
            const testObj = { message: "just a test" };
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            inPipe.write(testObj);
            inPipe.write(testObj);
            inPipe.write(testObj);
            chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")
                .get().values.reduce((prev, elem) => {
                if (prev) {
                    return prev;
                }
                if (elem.metricName === name + "_chunk_sizes_bytes_sum") {
                    return elem.value;
                }
            }, null)).equal(3);
        });
        it("should count an end", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_ends_total").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            outPipe.on("finish", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_ends_total")
                    .get().values[0].value).equal(1);
                done();
            });
            inPipe.end("Just a test");
        });
        it("should count a non-emit", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_non_emits_total").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            outPipe.on("finish", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_non_emits_total")
                    .get().values[0].value).equal(1);
                done();
            });
            inPipe.end();
        });
        it("should summarize the elapsed time", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_elapsed_time_ms").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            outPipe.on("finish", () => {
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_elapsed_time_ms")
                    .get().values[0].value).above(49);
                chai_1.expect(AutometricStreamPipe.register.getSingleMetric(name + "_elapsed_time_ms")
                    .get().values[0].value).below(55);
                done();
            });
            const chunks = ["just", "a", "test"];
            const intv = setInterval(() => {
                if (chunks.length) {
                    inPipe.write(chunks.shift());
                    return;
                }
                inPipe.end();
                clearInterval(intv);
            }, 50);
        });
    });
    describe("summary", () => {
        let AutometricStreamPipe;
        const name = "summaries";
        before(() => {
            AutometricStreamPipe = _1.createAutometricStreamPipe(name);
        });
        it("should measure execution duration", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_durations_ms").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            inPipe.write("test");
            setTimeout(() => {
                inPipe.end();
                setTimeout(() => {
                    const values = AutometricStreamPipe.register.getSingleMetric(name + "_durations_ms")
                        .get().values;
                    chai_1.expect(values[values.length - 2].value).greaterThan(999).lessThan(1010);
                    done();
                }, 10);
            }, 1000);
        });
        it("should measure execution duration after the first chunk", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_durations_ms").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            setTimeout(() => {
                inPipe.write("test");
            }, 100);
            setTimeout(() => {
                inPipe.end();
                setTimeout(() => {
                    const values = AutometricStreamPipe.register.getSingleMetric(name + "_durations_ms")
                        .get().values;
                    chai_1.expect(values[values.length - 2].value).greaterThan(890).lessThan(910);
                    done();
                }, 10);
            }, 1000);
        });
        it("should measure the number of chunks in one stream", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            let nums = 13;
            const intv = setInterval(() => {
                if (nums === 0) {
                    inPipe.end();
                    clearInterval(intv);
                    return;
                }
                nums -= 1;
                inPipe.write("chunk...");
            }, 10);
            outPipe.on("end", () => {
                const values = AutometricStreamPipe.register.getSingleMetric(name + "_chunk_sizes_bytes")
                    .get().values;
                chai_1.expect(values[values.length - 1].value).equal(13);
                done();
            });
        });
        it("should measure the throughput per stream", (done) => {
            AutometricStreamPipe.register.getSingleMetric(name + "_throughput_bytes").reset();
            const inPipe = new Pipe();
            const outPipe = new Pipe();
            const autometricStream = new AutometricStreamPipe();
            inPipe.pipe(autometricStream).pipe(outPipe);
            outPipe.on("data", () => {
                // do nothing, just consume...
            });
            for (let i = 0; i < 13; i += 1) {
                inPipe.write("chunk...");
            }
            inPipe.end();
            outPipe.on("end", () => {
                const values = AutometricStreamPipe.register.getSingleMetric(name + "_throughput_bytes")
                    .get().values;
                chai_1.expect(values[values.length - 2].value).equal(13 * 8);
                done();
            });
        });
    });
});
